import os
import time
import logging
import datetime
import signal
from subprocess import Popen, run

from moa.control import Manager

from .click_builder import ClickBuilder
from .click_communicator import ClickCommunicator, ClickCommunicatorException
from .click_writer import ClickWriter

class ClickManager(Manager):
    def __init__(self):
        self.interfaces_used = {}
        self.emulation = None
        self.graph = None
        self.log = logging.getLogger(__name__)

    def get_communicator(self):
        return ClickCommunicator()

    def set_emulation(self, emulation, graph):
        self.emulation = emulation
        self.graph = graph

    def serialize(self):
        serial_obj = {'type': 'ClickManager'}
        return serial_obj

    def get_interfaces_used(self):
        g = self.graph.get_graph()
        for node in g:
            if g.nodes[node]['type'] == 'interface' or g.nodes[node]['type'] == 'xdpinterface':
                iface = g.nodes[node]['info']
                self.interfaces_used[iface.id] = iface

        return self.interfaces_used

    def create_emulation(self):
        filename = self.emulation.filename
        m_graph = self.graph
        emulation_id = self.emulation.id
        writer = ClickWriter(filename)
        builder = ClickBuilder(writer, m_graph, str(emulation_id))

        self.emulation.interfaces = self.get_interfaces_used()

        writer.open()
        tm = datetime.datetime.utcnow()
        writer.write_comment("Emulation %s Created by %s at %s UTC" % (
            emulation_id,
            os.path.basename(__file__),
            tm.strftime("%Y-%m-%d %H:%M:%S")))

        builder.build_elements()
        builder.write_elements()
        builder.write_pipelines()
        writer.close()

        self.clear_elements()

        return True


    def clear_elements(self):
        g = self.graph.get_graph()
        for node in g:
            if 'elements' in g.nodes[node]:
                del g.nodes[node]['elements']
            g.nodes[node]['visited'] = False

    def start_emulation(self):
        # kill any existing socket
        run(["rm", "-f", self.emulation.socket_path])

        cmd = 'click -u %s %s' % (self.emulation.socket_path,
                                  self.emulation.filename)

        self.log.info('Running cmd: %s', cmd)
        process = Popen(cmd.split(), close_fds=True,
                        start_new_session=True, restore_signals=False)

        time.sleep(1) # give it sec to fail... Not sure this is necessary

        if process.poll():
            # Emulation failed to start, release lock and return error
            self.log.error(
                "Error starting click in user space. exit={}".format(process.poll()))
            return None

        self.log.info("Process with PID %d", process.pid)
        return process.pid

    def stop_by_pid(self, pid):
        self.log.info("Terminating process by PID %d", pid)
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError as oe:
            self.log.error("Error stopping pid %d - %s", pid, oe)
            return False

        time.sleep(1)
        # Check for death
        try:
            os.kill(pid, 0)
        except OSError:
            self.emulation.process = None
        else:
            # Still alive
            return False

        return self.emulation.process is None

    def stop_emulation(self):
        if isinstance(self.emulation.process, int):
            return self.stop_by_pid(self.emulation.process)

        self.log.info("Terminating process with PID %d", self.emulation.process.pid)

        self.emulation.process.terminate()

        time.sleep(1)   # give it a sec to close

        self.emulation.process.wait()
        # Check if the process died, os.kill sends a signal (probably should have a different name
        # but w/e) send signal 0 which will fail if the process is dead, do nothing otherwise.
        try:
            os.kill(self.emulation.process.pid, 0)
        except OSError:
            self.emulation.process = None
        else:
            # Still alive
            return False

        return self.emulation.process is None

    def update_click(self, communicator, node, key, value):
        ''' Update a given Click Elements configuration
            Wrapper for communicator.set_value to handle errors '''

        try:
            communicator.set_value(node, key, value)
        except ClickCommunicatorException as e:
            self.log.error(e)
            return False

        return True
